use log::{ debug, info, warn, error };
use dbus_crossroads::Crossroads;
use dbus::{
    message::MatchRule,
    channel::MatchingReceiver
};

const DBUS_NAME: &'static str = "org.ddnss.sfs.mc";

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    // env_logger initialization
    // If running with debug profile and `RUST_LOG` environment variable isn't set, set it to "debug"
    if cfg!(debug_assertions) && std::env::var_os(env_logger::DEFAULT_FILTER_ENV).is_none() {
        std::env::set_var(env_logger::DEFAULT_FILTER_ENV, "debug");
    }
    env_logger::init();

    debug!("PID: {}, User: {}", std::process::id(), std::env::var("USER").unwrap_or("N/A".into()));

    // New Crossroads instance
    let mut cr = Crossroads::new();

    // Insert object paths
    cr.insert("/", &[], ());

    // Connect with D-Bus
    let (resource, c) = dbus_tokio::connection::new_system_sync()?;
    debug!("D-Bus unique name: {}", c.unique_name());

    // Spawn a new task for the D-Bus resource
    tokio::spawn(async move {
        let err = resource.await;
        panic!("Lost connection to D-Bus: {}", err);
    });

    // Allow methods to be async
    cr.set_async_support(Some((c.clone(), Box::new(|x| { tokio::spawn(x); }))));

    // Aquire name on D-Bus
    match c.request_name(DBUS_NAME, false, true, false).await {
        Ok(_) => info!("Listening (D-Bus) on {}", DBUS_NAME),
        Err(err) => {
            error!("Couldn't request name on D-Bus: {}", err);
            std::process::exit(1);
        }
    };

    // Receive method calls
    c.start_receive(MatchRule::new_method_call(), Box::new(move |msg, conn| {
        cr.handle_message(msg, conn).unwrap_or_else(|_| warn!("Incoming message wasn't a method call"));
        // Return bool whether message should be handled only once (false) or another time (true)
        true
    }));

    // Will (afaik) be stabilized in Rust 1.48
    // core::future::pending::<()>().await;
    tokio::signal::unix::signal(tokio::signal::unix::SignalKind::terminate())?.recv().await;
    info!("Got SIGTERM, shutting down");
    Ok(())
}